#![allow(non_snake_case, dead_code)] // We're translating from someone else's JSON struct naming
//! Structs from the Knockout Forums JS API

use futures::stream::StreamExt;
use reqwest::Client;
use serde::{Deserialize, Serialize};
use std::cmp::Ordering;
use std::fs::File;
use std::io::{BufWriter, Write};
use std::sync::atomic::AtomicU64;
use std::sync::Arc;
use std::time::Duration;

#[derive(Serialize, Deserialize)]
pub struct PopularThread {
    pub id: u64,
}

impl PopularThread {
    pub async fn get() -> Result<Vec<PopularThread>, String> {
        let client = &Client::builder().build().unwrap();
        let url = "https://api.knockout.chat/v2/threads/popular";
        let mut iters: u64 = 0;
        loop {
            /////////////////////////////////////////////////////////////////
            // Connection Loop for 'in_thread'.
            // We use this to catch all the HTTP and deserilization errors we
            // might get on attempting to get the thread info.
            // Failing to get the thread info means we can't get any pages,
            // and will return an Error() as such.
            /////////////////////////////////////////////////////////////////
            iters = iters + 1; // Increment our iteration count.

            let resp: reqwest::Response;
            // Catch network errors and retry 10 times.
            match client.get(url.clone()).send().await {
                Ok(result) => {
                    resp = result;
                }
                Err(err) => {
                    if iters >= 10 {
                        // We can't connect for some deeper reason
                        // Return with the reason why.
                        return Err(err.to_string());
                    } else {
                        // Wait a bit for network conditions to improve and retry.
                        tokio::time::sleep(Duration::from_millis(iters * 25)).await;
                        continue;
                    }
                }
            }
            // Deal with HTTP response type.
            let resp_code = resp.status();
            match resp_code.is_client_error() {
                true => {
                    // This means it was an error on our part, probably a bad request
                    // Such as a thread that doesn't exist.
                    // Meaning we shouldn't retry, as opposed to other errors.
                    return Err(resp_code.to_string());
                }
                false => {}
            }
            // Get the text from our known good response, and deserialize it.
            let resp_body = resp.text().await.unwrap();
            let deser = serde_json::from_str(resp_body.as_str());
            match deser {
                Ok(data) => {
                    // Set our thread data and leave the loop.
                    return Ok(data);
                }
                Err(_err) => {
                    // We don't abort this retry, since the thread is valid,
                    // and we're probably just getting throttled.
                    // So keep trying until we get a response we can deserialize!

                    // DEBUG ERROR PRINTING
                    println!("Page: {}", url);
                    println!(
                        "\t THREAD DESER ERROR: {} \n\t TEXT: {}",
                        _err,
                        resp_body.as_str()
                    );
                     
                    // If we retry a LOT of times, something is seriously wrong, return an error.
                    if iters >= 25 {
                        return Err(_err.to_string());
                    } else {
                        // Backpressure, slow down request rate to alleviate throttling.
                        tokio::time::sleep(Duration::from_millis(iters * 10)).await;
                        continue; // From the top!
                    }
                }
            }
        }
    }
}

#[derive(Serialize, Deserialize)]
pub struct Subforum {
    pub id: u64,
    pub name: String,
    pub iconId: Option<u64>,
    pub createdAt: String,
    pub updatedAt: String,
    pub description: String,
}

#[derive(Serialize, Deserialize, Eq, PartialEq, Clone)]
pub enum Usergroup {
    BannedUser = 0,
    RegularUser = 1,
    GoldMember = 2,
    Moderator = 3,
    Admin = 4,
    Staff = 5,
    ModeratorInTraining = 6,
}

impl From<u64> for Usergroup {
    fn from(v: u64) -> Self {
        match v {
            0 => Usergroup::BannedUser,
            1 => Usergroup::RegularUser,
            2 => Usergroup::GoldMember,
            3 => Usergroup::Moderator,
            4 => Usergroup::Staff,
            5 => Usergroup::BannedUser,
            _ => Usergroup::RegularUser,
        }
    }
}

/// All groups which get Gold Member benefits
pub static GOLDMEMBER_GROUPS: &'static [Usergroup] = &[
    Usergroup::GoldMember,
    Usergroup::Moderator,
    Usergroup::Admin,
    Usergroup::Staff,
    Usergroup::ModeratorInTraining,
];

/// All groups with moderator permissions
static MODERATOR_GROUPS: &'static [Usergroup] = &[
    Usergroup::Moderator,
    Usergroup::Admin,
    Usergroup::Staff,
    Usergroup::ModeratorInTraining,
];

/// A Knockout User
///
/// Most of these fields were reverse-engineered from the API return JSON.
#[derive(Serialize, Deserialize, Eq, Clone)]
pub struct User {
    pub id: u64,
    pub username: Option<String>, // Potentially Null... 🙄😤
    pub Usergroup: Option<String>,
    pub avatarUrl: String,
    pub backgroundUrl: String,
    pub posts: u64,
    pub threads: u64,
    pub createdAt: String,
    pub updatedAt: String,
    pub banned: bool,
    pub isBanned: bool,
}

/// When ordering, only the user ID is taken into account.
impl Ord for User {
    fn cmp(&self, other: &Self) -> Ordering {
        self.id.cmp(&other.id)
    }
}
impl PartialOrd for User {
    fn partial_cmp(&self, other: &Self) -> Option<Ordering> {
        Some(self.cmp(other))
    }
}
impl PartialEq for User {
    fn eq(&self, other: &Self) -> bool {
        self.id == other.id
    }
}

/// Not implemented.
#[derive(Serialize, Deserialize, Eq, PartialEq)]
pub struct Ban {
    // Not implemented?
}

/// Not implemented.
#[derive(Serialize, Deserialize, Eq, PartialEq)]
pub struct Ratings {
    id: u64,
    rating: String,
    ratingId: u64,
    users: Vec<User>,
}

#[derive(Serialize, Deserialize, Debug, Eq, PartialEq)]
pub enum Rating {
    AGREE,
    DISAGREE,
    FUNNY,
    FRIENDLY,
    KAWAII,
    SAD,
    ARTISTIC,
    INFORMATIVE,
    IDEA,
    WINNER,
    GLASSES,
    LATE,
    DUMB,
    CITATION,
    OPTIMISTIC,
    ZING,
    YEET,
    UNKOWN,
}

impl From<&str> for Rating {
    fn from(str: &str) -> Self {
        match str {
            "agree" => Rating::AGREE,
            "disagree" => Rating::DISAGREE,
            "funny" => Rating::FUNNY,
            "friendly" => Rating::FRIENDLY,
            "kawaii" => Rating::KAWAII,
            "sad" => Rating::SAD,
            "artistic" => Rating::ARTISTIC,
            "informative" => Rating::INFORMATIVE,
            "idea" => Rating::IDEA,
            "winner" => Rating::WINNER,
            "glasses" => Rating::GLASSES,
            "late" => Rating::LATE,
            "dumb" => Rating::DUMB,
            "citation" => Rating::CITATION,
            "optimistic" => Rating::OPTIMISTIC,
            "zing" => Rating::ZING,
            "yeet" => Rating::YEET,
            _ => Rating::UNKOWN,
        }
    }
}

impl From<String> for Rating {
    fn from(in_string: String) -> Self {
        Rating::from(in_string.as_str())
    }
}

#[derive(Serialize, Deserialize, Eq)]
pub struct Post {
    pub id: u64,
    /// Global Thread ID
    pub thread: u64,
    /// Page Number within thread
    pub page: u64,
    pub content: String,
    /// ISO 8601
    pub createdAt: String,
    /// ISO 8601
    pub updatedAt: String,
    pub user: User,
    pub ratings: Option<Vec<Ratings>>,
    pub bans: Vec<Ban>,
    pub threadPostNumber: u64,
    pub countryName: Option<String>,
    ///ISO3166-1 Alpha 2
    pub countryCode: Option<String>,
    pub appName: Option<String>,
}

impl Ord for Post {
    fn cmp(&self, other: &Self) -> Ordering {
        self.threadPostNumber.cmp(&other.threadPostNumber)
    }
}

impl PartialOrd for Post {
    fn partial_cmp(&self, other: &Self) -> Option<Ordering> {
        Some(self.cmp(other))
    }
}

impl PartialEq for Post {
    fn eq(&self, other: &Self) -> bool {
        self.threadPostNumber == other.threadPostNumber
    }
}

#[derive(Serialize, Deserialize)]
pub struct Tags {
    // Not implemented?
}

#[derive(Serialize, Deserialize)]
pub struct Viewers {
    pub memberCount: u64,
    pub guestCount: u64,
}

/// A Knockout thread
///
/// Contains a huge vector of Posts.
#[derive(Serialize, Deserialize)]
pub struct Thread {
    id: u64,
    title: String,
    iconId: u64,
    subforumId: u64,

    /// ISO 8601 String
    createdAt: String,
    /// ISO 8601 String
    updatedAt: String,
    /// Option wrapping ISO 8601 String
    deletedAt: Option<String>,
    deleted: bool,
    locked: bool,
    pinned: bool,
    subscribed: bool,

    /// The most recent post in the thread
    lastPost: Post,
    postCount: u64,
    recentPostCount: u64,
    unreadPostCount: u64,
    readThreadUnreadPosts: u64,

    read: bool,
    hasRead: bool,
    hasSeenNoNewPosts: bool,

    subforum: Subforum,
    tags: Vec<Tags>,
    viewers: Viewers,

    totalPosts: u64,
    currentPage: u64,

    threadBackgroundUrl: Option<String>,
    threadBackgroundType: Option<String>,

    user: User,
    posts: Vec<Post>,
    readThreadLastSeen: Option<String>,
}

impl Thread {
    pub fn get_posts(self) -> Vec<Post> {
        self.posts
    }
    pub fn get_id(self) -> u64 {
        self.id
    }
    pub fn get_title(self) -> String {
        self.title
    }
    /////////////////////////////////////////////////////////////////////////////
    /// Gets the thread from the KO API.
    ///
    /// Posts are sorted in thread-posted order.
    /////////////////////////////////////////////////////////////////////////////
    pub async fn new_from_id(thread_id: u64) -> Result<Thread, String> {
        let client = &Client::builder().build().unwrap();
        let thread_url = format!("http://api.knockout.chat/v2/threads/{}", thread_id);
        let mut in_thread: Thread; // Will be allocated inside the loop.

        let mut iters: u64 = 0;
        loop {
            /////////////////////////////////////////////////////////////////
            // Connection Loop for 'in_thread'.
            // We use this to catch all the HTTP and deserilization errors we
            // might get on attempting to get the thread info.
            // Failing to get the thread info means we can't get any pages,
            // and will return an Error() as such.
            /////////////////////////////////////////////////////////////////
            iters = iters + 1; // Increment our iteration count.

            let get = client.get(thread_url.clone()).send().await;
            let resp: reqwest::Response;
            // Catch network errors and retry 10 times.
            match get {
                Ok(result) => {
                    resp = result;
                }
                Err(err) => {
                    if iters >= 10 {
                        // We can't connect for some deeper reason
                        // Return with the reason why.
                        return Err(err.to_string());
                    } else {
                        // Wait a bit for network conditions to improve and retry.
                        tokio::time::sleep(Duration::from_millis(iters * 25)).await;
                        continue;
                    }
                }
            }
            // Deal with HTTP response type.
            let resp_code = resp.status();
            match resp_code.is_client_error() {
                true => {
                    // This means it was an error on our part, probably a bad request
                    // Such as a thread that doesn't exist.
                    // Meaning we shouldn't retry, as opposed to other errors.
                    return Err(resp_code.to_string());
                }
                false => {}
            }
            // Get the text from our known good response, and deserialize it.
            let resp_body = resp.text().await.unwrap();
            let deser = serde_json::from_str(resp_body.as_str());
            match deser {
                Ok(data) => {
                    // Set our thread data and leave the loop.
                    in_thread = data;
                    break;
                }
                Err(_err) => {
                    // We don't abort this retry, since the thread is valid,
                    // and we're probably just getting throttled.
                    // So keep trying until we get a response we can deserialize!

                    /* // DEBUG ERROR PRINTING
                    println!("Page: {}", thread_url);
                    println!(
                        "\t THREAD DESER ERROR: {} \n\t TEXT: {}",
                        _err,
                        resp_body.as_str()
                    );
                     */
                    // If we retry a LOT of times, something is seriously wrong, return an error.
                    if iters >= 25 {
                        return Err(_err.to_string());
                    } else {
                        // Backpressure, slow down request rate to alleviate throttling.
                        tokio::time::sleep(Duration::from_millis(iters * 10)).await;
                        continue; // From the top!
                    }
                }
            }
        }

        // We got the thread information, so now we can beging scraping pages.
        let total_pages = in_thread.lastPost.page; // Total Pages in thread.
        let last_post_num = in_thread.lastPost.threadPostNumber as usize;
        let scrape_atomic = &Arc::new(AtomicU64::new(0));

        println!("Thread:\t{}\t\"{}\"", in_thread.id, in_thread.title);
        println!("Total Pages: {}", total_pages);

        // Builds our page URLs into a vector
        // This is later passed into an async unordered buffer executor.
        let mut urls: Vec<String> = Vec::with_capacity(total_pages as usize);
        for n in 1..=total_pages {
            urls.push(format!("{}/{}", thread_url.clone(), n));
        }

        // Collect all the pages into an unordered buffer.
        let fetches = futures::stream::iter(urls.into_iter().map(|url| {
            async move {
                let mut iters = 0;
                loop {
                    iters = iters + 1; // For backpressure.
                                       // Slow down a bit to not overwhelm the server.
                    tokio::time::sleep(Duration::from_millis(10)).await;

                    let thread_str: String; // Assigned in Ok client.get below.
                    loop {
                        match client.get(url.clone()).send().await {
                            Ok(res) => {
                                thread_str = res.text().await.unwrap();
                                break;
                            }
                            Err(err) => {
                                println!(
                                    "\t Failed to connect: {}\n\t  Reason: {}",
                                    url.clone(),
                                    err
                                );
                                continue;
                            }
                        }
                    }

                    let thread: Thread; // No initial value, assign inside following match.
                    match serde_json::from_str(thread_str.as_str()) {
                        Ok(deser) => {
                            thread = deser;
                        }
                        Err(_err) => {
                            if _err.to_string().contains("invalid type") {
                                println!("Page: {}", url);
                                let path = format!(
                                    "threads/{}-{}.err.json",
                                    thread_id,
                                    url.clone().split("/").last().unwrap()
                                );
                                println!("\t DESER ERROR: {}", _err);
                                println!("\t Wrote error to {}", path);
                                {
                                    // Scope specifically for file handling.
                                    let file = File::create(path).unwrap();
                                    let mut fbuf = BufWriter::new(file);
                                    writeln!(fbuf, "{}", thread_str.as_str()).unwrap();
                                    write!(fbuf, "{}", _err).unwrap();
                                }
                                return Err("Deserialization Errors".to_string());
                            }
                            if iters >= 25 {
                                return Err(_err.to_string());
                            }
                            // Probably just overloaded the server, hold up a sec.
                            tokio::time::sleep(Duration::from_millis(iters * 25)).await;
                            continue; // Restart on this request.
                        }
                    }
                    // Increment our progress
                    let mut progress = scrape_atomic.load(std::sync::atomic::Ordering::Acquire);
                    progress = progress + 1;
                    scrape_atomic.store(progress, std::sync::atomic::Ordering::Release);
                    // Build a list of users.
                    //let users = build_users(thread.posts);
                    // Display our progress
                    println!(
                        "Processed:... {:width$}/{page} | Page: {realpage:width$}",
                        progress,
                        width = format!("{}", last_post_num).chars().count(),
                        page = thread.lastPost.page,
                        realpage = thread.posts.last().unwrap().page,
                    );
                    break Ok(thread.posts);
                }
            }
        }))
        // KO accepts up to 25 connections, we already utilized 1 for the thread
        // So we're going to cap the maximum requests here to 24.
        .buffer_unordered((total_pages as usize).min(24).max(1))
        .collect::<Vec<Result<Vec<Post>, String>>>()
        .await;

        // Merge our unordered buffer into a straight Vec.
        let mut post_buffer: Vec<Post> = Vec::with_capacity(last_post_num);
        for vec in fetches.into_iter() {
            match vec {
                Ok(mut posts) => {
                    post_buffer.append(&mut posts);
                }
                // On first encounter of an error, send it back to the calling function
                Err(err) => {
                    return Err(err);
                }
            }
        }
        // Sort our Post and UserDocument Vecs.
        post_buffer.sort();
        // Move all posts from all pages into the original thread struct.
        in_thread.posts = post_buffer;
        Ok(in_thread)
    }
}

/// Takes a vector of posts and returns all the users who posted
fn build_users(posts: Vec<Post>) -> Vec<User> {
    let mut users: Vec<User> = Vec::new();
    for i in posts.iter() {
        users.push(i.user.clone());
    }
    users
}
